#include <linux/irq.h>
#include <linux/gfp.h>
#include <linux/seq_file.h>
#include <linux/interrupt.h>
#include <linux/sysdev.h>
#include <mach/gpio.h>
#include <plat/gpio.h>

#define APMASK_OFFSET   0x9C  /* The APMASK register is in the Marvell's GPIO Control Block */

#define FPGA_REGS_BASE TS47XX_FPGA_VIRT_BASE
#define FPGA_REG(x) (*((volatile u16 *)(FPGA_REGS_BASE + (x))))
#define FPGA_IRQ_OFFSET       0x2C     /* IRQ cause */
#define FPGA_IRQ_MASK_OFFSET  0x2E     /* IRQ mask (1=masked)*/

#define MIN_FPGA_IRQ    64
#define MAX_FPGA_IRQ    (MIN_FPGA_IRQ + 15)
#define IS_BAD_FPGA_IRQ(x) ((x) < MIN_FPGA_IRQ || (x) > MAX_FPGA_IRQ)


static void mux_irq_handler(unsigned int irq, struct irq_desc *desc)
{
   unsigned int bit, n;
   unsigned short fpga_irq_reg;

   /*
    * When the FPGA asserts its IRQ line, we end up in this handler.  Now we need
    * to read the FPGA's IRQ register to determine what peripheral(s) need to be serviced.
    */

   GPIO_REG(GEDR_OFFSET) = BIT(27);          /* Clear the status bit by writing 1 */

   /*    Read the IRQ register from the FPGA...
    *
    *    BIT 0: XUART            (IRQ #64)
    *    BIT 1: CAN              (IRQ #65)
    *    BIT 2: CAN2             (IRQ #66)
    *    BIT 3: PC/104 IRQ 5     (IRQ #67)
    *    BIT 4: PC/104 IRQ 6     (IRQ #68)
    *    BIT 5: PC/104 IRQ 7     (IRQ #69)
    *    BITS 6-15: Reserved     (IRQ #70-80)
    */

   while((fpga_irq_reg = FPGA_REG(FPGA_IRQ_OFFSET))) {
      /* printk("fpga_irq_reg: 0x%04X\n", fpga_irq_reg); */
      for(n=0,bit=1; bit < 0x8000; bit <<= 1, n++)  {
         if(fpga_irq_reg & bit)
            generic_handle_irq(MIN_FPGA_IRQ + n);
      }
   }
}


static void unmask_muxed_irq(unsigned int irq)
{
   int bit = irq - MIN_FPGA_IRQ;
   if IS_BAD_FPGA_IRQ(irq)
      return;

   FPGA_REG(FPGA_IRQ_MASK_OFFSET) &= ~BIT(bit);
}

static void mask_muxed_irq(unsigned int irq)
{
   int bit = irq - MIN_FPGA_IRQ;
   if IS_BAD_FPGA_IRQ(irq)
      return;

   FPGA_REG(FPGA_IRQ_MASK_OFFSET) |= BIT(bit);
}

static void ack_muxed_irq(unsigned int irq)
{
#if (0)
   /* temp hack for test... */
   int bit = irq - MIN_FPGA_IRQ;
   if IS_BAD_FPGA_IRQ(irq)
      return;

   printk("ack %d\n", irq);
   FPGA_REG(0x10) = 0;
#endif
}

static struct irq_chip muxed_irq_chip = {
   .name    = "MUXIRQ",
   .mask    = mask_muxed_irq,
   .unmask  = unmask_muxed_irq,
   .ack     = ack_muxed_irq
};


void __init ts4700_init_mux_irq(void)
{
   int n;

   /*
    * On the TS-4700, the FPGA interrupt is connected to MFP_27.  This pin must
    * be configured as a GPIO pin (this is done in ts4700.c -- look for "GPIO27_GPIO").
    */

   FPGA_REG(FPGA_IRQ_MASK_OFFSET) = 0xFFFF;  /* mask all */

   GPIO_REG(GPDR_OFFSET) &= ~BIT(27);     /* Make sure GPIO #27 is an input pin */
   GPIO_REG(GRER_OFFSET) |=  BIT(27);     /* Enable rising-edge-detect on this input */
   GPIO_REG(GFER_OFFSET) &= ~BIT(27);     /* Disable falling-edge-detect on this input */
   GPIO_REG(GEDR_OFFSET)  =  BIT(27);     /* Ensure the status bit is clear by writing 1 */


   for(n=MIN_FPGA_IRQ; n <= MAX_FPGA_IRQ; n++)
   {
      set_irq_chip(n, &muxed_irq_chip);
      set_irq_handler(n, handle_edge_irq);
      set_irq_flags(n, IRQF_VALID);
   }


   /* Install mux handler */
   set_irq_chained_handler(IRQ_PXA168_GPIOX, mux_irq_handler);

   GPIO_REG(APMASK_OFFSET) |= BIT(27);    /* Unmask GPIO #27 */

}

